                .LF     dice6.lst
;------------------------------------------------------------------------
;
;  dice6.asm
;
;  Roll up to 6 dice on the MPF-1
;
;  Author : San Bergmans
;  Source : www.sbprojects.com
;  Date   : 20-12-2005
;
;------------------------------------------------------------------------
;
;  This program simulates 6 dice. You roll them by pressing the GO key.
;  The rolling is simulated by some animated patterns on the display for
;  as long as you hold the GO key down.
;  After the roll you can lock any of the dice by pressing the digit
;  keys 1..6, where 1 represents the left most die and 6 the right most.
;  Every time you press a certain digit you toggle the lock flag for that
;  die. A locked die is indicated by a dot on its right hand side.
;  Locked dice will not roll the next time you press GO again.
;  You can clear all the locks at once by pressing the DEL key.
;
;------------------------------------------------------------------------

                .CR     Z80             Yeah! We're programming the Z80!
                .OR     01800H          RAM starts here, so are we
                .TF     dice6.hex,HEX   Create a plain hex dump file

;------------------------------------------------------------------------
;  Constants
;------------------------------------------------------------------------

ROLL_DELAY      .EQ     6               Animation speed factor (x 10ms)

PRIME1          .EQ     1               Some prime numbers used to
PRIME2          .EQ     3                make sure none of the rolls
PRIME3          .EQ     7                 will be predictable
PRIME4          .EQ     11              Prime numbers will ensure we
PRIME5          .EQ     13               won't skip any of the values
PRIME6          .EQ     17

GO_KEY          .EQ     012H            GO key  (SCAN)
DEL_KEY         .EQ     017H            DEL key (SCAN)

SA              .EQ     00001000B       Segment A
SB              .EQ     00010000B       Segment B
SC              .EQ     00100000B       Segment C
SD              .EQ     10000000B       Segment D
SE              .EQ     00000001B       Segment E
SF              .EQ     00000100B       Segment F
SG              .EQ     00000010B       Segment G
DP              .EQ     01000000B       Decimal point

;------------------------------------------------------------------------
;  Monitor routines
;------------------------------------------------------------------------

SCAN            .EQ     005FEH          Scan display until a key pressed
KEYMAP          .EQ     0061DH          Unofficial keymap routine
SCAN1           .EQ     00624H          Scan display once
HEX7            .EQ     00689H          Convert digit to 7-segement
HEX7SG          .EQ     00678H          Convert 2 hex digits to 7-segment
TONE1K          .EQ     005DEH          Sound a 1kHz tone during (HL)

;------------------------------------------------------------------------
;  START
;  Show the "Splash screen"
;------------------------------------------------------------------------

DICE            LD      IX,SPLASH       Show initial message (*DICE*)

.LOOP           LD      A,(ROLL_INDEX)  Get animation index value
                LD      E,A              and use it as index in
                LD      D,0              rolling pattern
                LD      HL,ROLL_CW      Roll right digit clockwise
                ADD     HL,DE
                LD      A,(HL)
                LD      (SPLASH),A
                LD      HL,ROLL_CCW     Roll left digit counter
                ADD     HL,DE            clockwise
                LD      A,(HL)
                LD      (SPLASH+5),A

                CALL    SCAN1_SEED      Scan display once
                JR      C,.LOOP         Do until a key is pressed!

                LD      HL,LAST_KEY     Same key down? GO key may still
                CP      A,(HL)           be down after starting the
                JR      Z,.LOOP           program!

                LD      (HL),A          Save new last key
                CP      A,GO_KEY        The only key allowed now is GO
                JR      NZ,.LOOP        It's not the GO key!

                LD      IX,DSP_BFFR     From now on show dice buffer
                JR      ROLL            Roll the dice now!

;------------------------------------------------------------------------
;  MAIN
;  The current dice are shown and we are waiting for a key to be pressed
;------------------------------------------------------------------------

MAIN            LD      IX,DSP_BFFR     From now on show dice buffer

.LOOP           CALL    SCAN1_SEED      Display current dice
                JR      C,.LOOP         No key is pressed!

                CP      A,DEL_KEY       Was the DEL key pressed?
                JR      Z,CLEAR_ALL     Yes! Clear all locks

                CP      A,6+1           Was a digit 1..6 pressed?
                JR      C,LOCK_DIE      It was! Toggle lock for that die

                CP      A,GO_KEY        Was the GO key pressed?
                JP      Z,ROLL          Yes! Roll the free dice

                JR      .LOOP           Repeat this indefinitely

;------------------------------------------------------------------------
;  ROLL
;  Roll the dice for as long as the GO key remains pressed while showing
;  an animated rolling pattern on the displays of the rolling dice.
;------------------------------------------------------------------------

ROLL            CALL    SHOW_ROLL       Handle rolling animation
                CALL    SCAN1_SEED      Scan display once
                JR      NC,ROLL         Key still down!

                CALL    ROLL_DICE       Calculate the random rolls
                CALL    SHOW_DICE        and display them

                JR      MAIN            Return to main routine now!

;------------------------------------------------------------------------
;  LOCK_DIE
;  A digit key is pressed. Toggle the lock flag of the corresponding die.
;------------------------------------------------------------------------

LOCK_DIE        LD      HL,LAST_KEY
                CP      A,(HL)          Same key as last time?
                JR      Z,.EXIT         Yes! Accept key only once

                LD      (HL),A          Save new last key
                AND     A,A             Ignore the 0 key
                JR      Z,.EXIT         It was the 0 key!

                LD      HL,DIE1-1       Prepare die pointer
                LD      E,A              (Mind the offset of -1!, this is
                LD      D,0               because the keys are numbered
                ADD     HL,DE              form 1 to 6)
                LD      A,10000000B     Toggle b7 of this die
                XOR     A,(HL)
                LD      (HL),A

                CALL    SHOW_DICE       Show new status

.EXIT           JR      MAIN            And return to main loop

;------------------------------------------------------------------------
;  CLEAR_ALL
;  The DEL key was pressed, clear all locks now.
;------------------------------------------------------------------------

CLEAR_ALL       LD      HL,DIE1         Clear all lock bits
                LD      B,6              of all 6 dice

.LOOP           RES     7,(HL)          Clear b7. Nice Z80 instruction!
                INC     HL
                DJNZ    .LOOP           Do all dice!

                CALL    SHOW_DICE       Show new status

                JR      MAIN

;------------------------------------------------------------------------
;  SHOW_DICE
;  Convert all 6 dice to 7-segment patterns. Also display a dot to right
;  of the die if it is locked.
;------------------------------------------------------------------------

SHOW_DICE       LD      HL,DIE6         Prepare to load all dice
                LD      DE,DSP_BFFR
                LD      B,6             Do 6 dice

.LOOP           LD      A,(HL)          Get this die's value
                LD      C,A             Save lock flag (b7) of this die
                CALL    HEX7            Convert die to 7-segment
                RL      C               Is lock flag set for this die?
                JR      NC,.NOLOCK      Nope!
                OR      A,DP            Set decimal point for locked die

.NOLOCK         EX      DE,HL           Save pattern in display buffer
                LD      (HL),A
                EX      DE,HL
                DEC     HL              Previous die (going backwards!)
                INC     DE              Next display buffer location
                DJNZ    .LOOP           Do all 6 dice!
                RET

;------------------------------------------------------------------------
;  SCAN1_SEED
;  Increment the seeds of all 6 dice with different step size. Also the
;  animation index is incremented. Then scan the display once. If a key
;  appears to be down it is translated into an internal key code which
;  makes processing the digit keys easier.
;  If no key is down at the end of this routine the last key flag is
;  cleared.
;------------------------------------------------------------------------

SCAN1_SEED      LD      HL,SEED1        Increment the seeds using 6
                LD      A,(HL)           different prime numbers to
                ADD     A,PRIME1          make the dice's rolls quite
                LD      (HL),A             unpredictable

                INC     HL
                LD      A,(HL)
                ADD     A,PRIME2
                LD      (HL),A

                INC     HL
                LD      A,(HL)
                ADD     A,PRIME3
                LD      (HL),A

                INC     HL
                LD      A,(HL)
                ADD     A,PRIME4
                LD      (HL),A

                INC     HL
                LD      A,(HL)
                ADD     A,PRIME5
                LD      (HL),A

                INC     HL
                LD      A,(HL)
                ADD     A,PRIME6
                LD      (HL),A

                LD      HL,ROLL_SPEED   Decrement animation delay
                DEC     (HL)             until it's 0
                JR      NZ,.SKIP        Don't reload until it is 0!
                LD      (HL),ROLL_DELAY Reload with delay value
                INC     HL              Point to rolling index
                DEC     (HL)
                JP      P,.SKIP         Don't reload while positive!
                LD      (HL),5          Reload to reduce to 6 steps
.SKIP
                CALL    SCAN1           Now scan the display
                JP      NC,KEYMAP       A key is down! Convert it!
                LD      A,0FFH          Clear last key flag
                LD      (LAST_KEY),A
                RET

;------------------------------------------------------------------------
;  SHOW_ROLL
;  Show the rolling animation of all free dice. Odd dice roll clockwise,
;  while the even dice are rolling counter clockwise.
;------------------------------------------------------------------------

SHOW_ROLL       LD      HL,ROLL_INDEX   Point to rolling index
                LD      A,(HL)          Get rolling index value
                LD      E,A              and use it as index in
                LD      D,0              rolling pattern
                LD      HL,ROLL_CW      Get CW pattern
                ADD     HL,DE
                LD      A,(HL)
                LD      B,A             Save this one
                LD      HL,ROLL_CCW     Also get CCW pattern
                ADD     HL,DE
                LD      A,(HL)
                LD      C,A             Save this one too

                LD      HL,DIE6         Do this for all dice
                LD      D,6
                LD      IY,DSP_BFFR     Point to display buffer

.LOOP           LD      A,(HL)          Get lock flag of this die
                RLA                      (lock flag is in b7)
                JR      C,.LOCKED       Don't animate when locked!
                LD      A,D             See if display is even or odd
                RRA
                JR      NC,.EVEN        This one's even!
                LD      A,B             Get clockwise pattern
                JR      .ODD
.EVEN           LD      A,C             Get counter clockwise pattern
.ODD            LD      (IY),A          Save pattern in display buffer

.LOCKED         INC     IY              Point to next display buffer
                DEC     HL              Point to next die
                DEC     D               Decrement die counter
                JR      NZ,.LOOP        Do all 6 of them!
                RET

;------------------------------------------------------------------------
;  ROLL_DICE
;  Calculate up to 6 new random numbers between 1 and 6. Of course only
;  for unlocked dice!
;------------------------------------------------------------------------

ROLL_DICE       LD      DE,5            Roll all dice in turn (backwards)
.ROLL           LD      HL,DIE1         Point to all die rolls
                ADD     HL,DE
                LD      A,(HL)          See if this die must roll
                RLA
                JR      C,.NOROLL       It shouldn't if b7 is set!

                CALL    ROLL_DIE        Calculate random number
                LD      (HL),A          Save new roll

.NOROLL         DEC     E               Roll them all (including 0!)
                JP      P,.ROLL          so we can't use DJNZ here
                RET

;------------------------------------------------------------------------
;  ROLL_DIE
;  Calculate random number for selected die. DE points to die properties.
;------------------------------------------------------------------------

ROLL_DIE        PUSH    DE              Save affected registers
                PUSH    HL

                LD      HL,SEED1        Use die number as offset in
                ADD     HL,DE            seed table
                LD      A,(HL)          Get this die's seed

                AND     A,A             Is seed 0?
                JR      Z,.ZERO         Yes! 0 not handled by generator!

                AND     A,#10111000B    Use seed to calculate random
                SCF                      number
                JP      PO,.SF
                CCF
.SF             LD      A,(HL)
                RLA
.ZERO           LD      (HL),A          Save new seed value

                LD      L,A             Multiply seed by 6 now to get
                LD      H,0              a value from 0..5
                ADD     HL,HL             (Hint: ADD HL,HL is the same
                LD      D,H                 as multiplying HL by 2)
                LD      E,L
                ADD     HL,HL
                ADD     HL,DE
                LD      A,H             Use only integer part
                INC     A               Make the value range from 1..6

                POP     HL              Restore affected registers
                POP     DE

                RET

;------------------------------------------------------------------------
;  Roll animation tables for counter clockwise and clockwise rolling
;  patterns.
;------------------------------------------------------------------------

ROLL_CCW        .DA     #SA+SB          Always having 2 segments on at a
                .DA     #SB+SC           time
                .DA     #SC+SD
                .DA     #SD+SE
                .DA     #SE+SF
                .DA     #SF+SA

ROLL_CW         .DA     #SA+SF
                .DA     #SF+SE
                .DA     #SE+SD
                .DA     #SD+SC
                .DA     #SC+SB
                .DA     #SB+SA

;------------------------------------------------------------------------
;  RAM variables
;------------------------------------------------------------------------

LAST_KEY        .DA     #GO_KEY         Last pressed key

SEED1           .RF     1               Random seed for die 1
SEED2           .RF     1               Random seed for die 2
SEED3           .RF     1               Random seed for die 3
SEED4           .RF     1               Random seed for die 4
SEED5           .RF     1               Random seed for die 5
SEED6           .RF     1               Random seed for die 6

DIE1            .DA     #1              Die value 1 (left most die)
DIE2            .DA     #1
DIE3            .DA     #1
DIE4            .DA     #1
DIE5            .DA     #1
DIE6            .DA     #1

ROLL_SPEED      .DA     #ROLL_DELAY     Speed of roll animation
ROLL_INDEX      .DA     #0              Rolling index for animation

;------------------------------------------------------------------------

DSP_BFFR        .DA     #0              Right most digit of display
                .DA     #0
                .DA     #0
                .DA     #0
                .DA     #0
                .DA     #0

;------------------------------------------------------------------------

SPLASH          .DA     #0              "Splash screen"
                .DA     #SA+SD+SE+SF+SG     E
                .DA     #SA+SD+SE+SF        C
                .DA     #SE+SF              I
                .DA     #SB+SC+SD+SE+SG     D
                .DA     #0

;------------------------------------------------------------------------

                .LI     OFF
